#include "lscvcudakernels.h"
#include "cudamath.h"
#include <cutil_inline.h>
#include "cudautils.h"


__device__ __constant__ float invSigma[1024];

__device__ inline float ceilSquareRoot(uint x) {
	float a=sqrtf(x);
	//return round(a)*round(a)==x?round(a):ceil(a);
	return round(a)*round(a)==x?round(a):a;
}

__device__ float kstar_dev(float h2,float e,float C1, float C2) {
	float k;
	float k2;

	k=exp(-e/(2.0f*h2))/C1;
	k2=exp(-e/(4.0f*h2))/C2;

	return k2-2.0f*k;
}

template <unsigned int blockSize> __global__ void reduceKstarKernel(float *g_idata, float *g_odata, unsigned int n,float C1, float C2, float minh, float step) {
	extern __shared__ float sdata[];
	unsigned int tid = threadIdx.x;
	sdata[tid] = 0.0f;

	unsigned int i = blockIdx.x*(blockSize*2) + tid;
	unsigned int gridSize = blockSize*2*gridDim.x;
	float h2=minh+step*blockIdx.y;
	h2=h2*h2;

	while (i+blockSize < n) { sdata[tid] += kstar_dev(h2,g_idata[i],C1,C2) + kstar_dev(h2,g_idata[i+blockSize],C1,C2);  i += gridSize;  }
	if (i < n ) { sdata[tid] += kstar_dev(h2,g_idata[i],C1,C2); }

	__syncthreads();

	if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
	if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
	if (blockSize >= 128) { if (tid < 64)  { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }

	if (tid < 32) {
		volatile float* smem = sdata;
		if (blockSize >=  64) smem[tid] += smem[tid + 32];
		if (blockSize >=  32) smem[tid] += smem[tid + 16];
		if (blockSize >=  16) smem[tid] += smem[tid +  8];
		if (blockSize >=    8) smem[tid] += smem[tid +  4];
		if (blockSize >=    4) smem[tid] += smem[tid +  2];
		if (blockSize >=    2) smem[tid] += smem[tid +  1];
	}

	if (tid == 0) g_odata[blockIdx.x+gridDim.x*blockIdx.y] = sdata[0];
}

template <unsigned int blockSize> __global__ void reduceMultirowKernel(float *g_idata, float *g_odata, unsigned int n) {
	extern __shared__ float sdata[];
	unsigned int tid = threadIdx.x;
	sdata[tid] = 0.0f;

	unsigned int i = blockIdx.x*(blockSize*2) + tid;
	unsigned int gridSize = blockSize*2*gridDim.x;
	unsigned int deltaY=n*blockIdx.y;

	while (i+blockSize < n) { sdata[tid] += g_idata[i+deltaY] + g_idata[i+blockSize+deltaY];  i += gridSize;  }
	if (i < n ) { sdata[tid] += g_idata[i+deltaY]; }

	__syncthreads();

	if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
	if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
	if (blockSize >= 128) { if (tid < 64)  { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }

	if (tid < 32) {
		volatile float* smem = sdata;
		if (blockSize >=  64) smem[tid] += smem[tid + 32];
		if (blockSize >=  32) smem[tid] += smem[tid + 16];
		if (blockSize >=  16) smem[tid] += smem[tid +  8];
		if (blockSize >=    8) smem[tid] += smem[tid +  4];
		if (blockSize >=    4) smem[tid] += smem[tid +  2];
		if (blockSize >=    2) smem[tid] += smem[tid +  1];
	}

	if (tid == 0) g_odata[blockIdx.x+gridDim.x*blockIdx.y] = sdata[0];
}

__global__ void finalizeCalc(float *g_idata, float *g_odata, uint values, float n,float d, float minh, float step,float C2) {
	uint tid=threadIdx.x+blockIdx.x*blockDim.x;

	if (tid<values) {
		float sum=g_idata[tid];
		float h=minh+tid*step;
		g_odata[tid]=(2*sum/((float)(n*n))+1.0f/(n*C2))/pow(h,d);
		//g_odata[tid]=g_idata[tid];
	}
}

template <unsigned int blockSize> __host__ void reduceAndCalcGH(float *d_idata, float C1, float C2, float minh, float step, uint values, float *results,uint n,uint d) {
	//WARNING Values nie może być większe niż 65535!
	//Allocate buffer
	uint tempBufferWidth=min(65535,intDivCeil(n*(n-1)/2,blockSize*2)); //WARNING 65535 czy 655536? sprawdzić
	float *d_tempBuffer;
	cutilSafeCall(cudaMalloc((void**)&d_tempBuffer,sizeof(float)*tempBufferWidth*values));

	//First step - reduce the input data to the tempBuffer
	dim3 grid(tempBufferWidth,values);
	reduceKstarKernel<blockSize><<<grid,blockSize,sizeof(float)*blockSize>>>(d_idata,d_tempBuffer,n*(n-1)/2,C1,C2,minh,step);
	//CUT_CHECK_ERROR("Kernel execution failed");

	//Second step - use ping-pong technique to reduce the tempBuffer to a single value
	float *d_out=d_tempBuffer;
	float *d_in;
	float *d_tempBuffer2=NULL;
	if (tempBufferWidth>1) {
		int size=intDivCeil(tempBufferWidth,2*blockSize);
		cutilSafeCall(cudaMalloc((void**)&d_tempBuffer2,sizeof(float)*size*values));
	}
	d_in=d_tempBuffer2;

	float *temp;
	uint oldTempBufferWidth;

	while (tempBufferWidth>1) {
		temp=d_out;
		d_out=d_in;
		d_in=temp;

		oldTempBufferWidth=tempBufferWidth;
		tempBufferWidth=intDivCeil(tempBufferWidth,2*blockSize);

		dim3 grid(tempBufferWidth,values);
		reduceMultirowKernel<blockSize><<<grid,blockSize,sizeof(float)*blockSize>>>(d_in,d_out,oldTempBufferWidth);
	}


	finalizeCalc<<<intDivCeil(values,blockSize),blockSize>>>(d_out,results,values,n,d,minh,step,C2);


	cutilSafeCall(cudaFree(d_tempBuffer));
	if (d_tempBuffer2!=NULL) {
		cutilSafeCall(cudaFree(d_tempBuffer2));
	}

}

//Calculationg exponents array on GPU

template <int side,int d> __global__ void findExponentsDevSmall(float *data, float* out, int n) {
	extern __shared__ float base[];

	float* Mi=base;
	float* Mj=base+side*side;

	int bx=blockIdx.x+blockIdx.y*gridDim.x;

	int i=ceil((ceilSquareRoot((bx<<3)+9)-3.0f)/2.0f);
	int j=bx-((i*(i+1))>>1);

	int tx=threadIdx.x%side;
	int ty=threadIdx.x/side;

	int iBegin=i*side;
	int jBegin=j*side;

	float part=0.0f;
	float res=0.0f;

	Mj[ty*side+tx]=data[min(jBegin+tx,n-1)+min(ty,d-1)*n]; 
	Mi[ty*side+tx]=data[min(iBegin+tx,n-1)+min(ty,d-1)*n]; 

	__syncthreads();

	if ((jBegin+tx)<(iBegin+ty)) {
		//if ((jBegin+tx)<n && (iBegin+tx)<n ) {
			for (int p=0;p<d;p++) {
				part=0.0f;
				for (int k=0;k<d;k++) {
					part+=invSigma[k*d+p]*(Mi[k*side+tx]-Mj[k*side+ty]);
				}
				res+=part*(Mi[p*side+tx]-Mj[p*side+ty]);
			}
		//}

		int correction=(side*(side+1)*(i)+(j==i)*(2*side-ty+1)*ty)/2;
		out[threadIdx.x+bx*blockDim.x-correction]=res;
	}
}

float calcH(HAlgorithm<float>* a,float *data, int n, int d,float step) {
//float h_gpu(uint n, uint d, float step, float* data,double &timeExp,double &timeGh, double &timeTotal, uint &steps) {

	TIMESTAMP totalStart,totalEnd, cpuGpuStart,cpuGpuEnd,gpuCpuStart,gpuCpuEnd;
	
	totalStart=getTimer();

	float *cov=covarianceMatrix(data,d,n);

	float detCov;
	float *invCov;

	invertMatrixAndFindDeterminant(d,cov,&invCov,&detCov);

	float C1=pow(2*M_PI,(float)d/2.0f)*sqrt(detCov);
	float C2=pow(4*M_PI,(float)d/2.0f)*sqrt(detCov);

	cpuGpuStart=getTimer();
	
	cutilSafeCall(cudaMemcpyToSymbol(invSigma,invCov,d*d*sizeof(float)));

	float *d_exponents;
	cutilSafeCall(cudaMalloc((void**)&d_exponents,sizeof(float)*n*(n-1)/2));

	float *d_data;
	cutilSafeCall(cudaMalloc((void**)&d_data,sizeof(float)*n*d));
	cutilSafeCall(cudaMemcpy(d_data,data,sizeof(float)*n*d,cudaMemcpyHostToDevice));

	cpuGpuEnd=getTimer();
	
	/*int blocks=(n/16)*((n/16)+1)/2;
	int blocksY=intDivCeil(blocks,65535);

	blocks=intDivCeil(blocks,blocksY);*/
	
	
	int v1=intDivCeil(n,16);
	int v2=v1+1;
	
	if ((v1&1)==0) v1>>=1; else v2>>=1;
	
	int blocks=v1;
	int blocksY=v2;

	dim3 grid=dim3(blocks,blocksY);
	switch (d) {
		case 1:	findExponentsDevSmall<16,1><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 2:	findExponentsDevSmall<16,2><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 3:	findExponentsDevSmall<16,3><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 4:	findExponentsDevSmall<16,4><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 5:	findExponentsDevSmall<16,5><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 6:	findExponentsDevSmall<16,6><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 7:	findExponentsDevSmall<16,7><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 8:	findExponentsDevSmall<16,8><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 9:	findExponentsDevSmall<16,9><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 10:	findExponentsDevSmall<16,10><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 11:	findExponentsDevSmall<16,11><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 12:	findExponentsDevSmall<16,12><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 13:	findExponentsDevSmall<16,13><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 14:	findExponentsDevSmall<16,14><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 15:	findExponentsDevSmall<16,15><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
		case 16:	findExponentsDevSmall<16,16><<<grid,16*16,16*16*2*sizeof(float)>>>(d_data, d_exponents, n);break;
	}
	CUT_CHECK_ERROR("Exponents kernel execution failed");

	//cutilSafeCall(cudaThreadSynchronize());

	float rk_mik2_inv=d*d*pow(2,d)*pow(M_PI,(float)d/2.0f);
	float rf_inv=(pow(2,d+2)*pow(M_PI,(float)d/2.0f))/((float)(d*(d+2)));

	float h0=pow((float)d*rf_inv/(rk_mik2_inv*(float)n),1/(d+4.0f));

	float h0_min=h0/4.0f;
	float h0_max=h0*4.0f;


	//int steps=ceil((h0_max-h0_min)/step);
	int steps=150;//WARNING rozwiązanie tymczasowe, ignoruje inne procedury
	step=(h0_max-h0_min)/steps;

	float *d_results;

	cutilSafeCall(cudaMalloc((void**)&d_results,sizeof(float)*steps));

	reduceAndCalcGH<512>(d_exponents, C1, C2, h0_min, step, steps, d_results,n,d);

	float* results=new float[steps];

	gpuCpuStart=getTimer();
	cutilSafeCall(cudaMemcpy(results,d_results,sizeof(float)*steps,cudaMemcpyDeviceToHost));
	gpuCpuEnd=getTimer();
	float ghest_min=results[0];
	uint ind=0;


	for (int i=1;i<steps;i++) {
		//printf("%e \n",results[i]);
		if (ghest_min>results[i]) {
			ghest_min=results[i];
			ind=i;
		}
	}

	float hmin=(float)ind*step+h0_min;

	

	cutilSafeCall(cudaFree(d_results));
	cutilSafeCall(cudaFree(d_exponents));
	cutilSafeCall(cudaFree(d_data));
	delete []results;
	delete []cov;
	delete []invCov;
	totalEnd=getTimer();
	
	a->addMeasure("LSCV_CUDA_CPU->GPU_TRAN",cpuGpuEnd-cpuGpuStart);
	a->addMeasure("LSCV_CUDA_GPU->CPU_TRAN",gpuCpuEnd-gpuCpuStart);
	a->addMeasure("LSCV_CUDA_STEPS",steps);
	a->addMeasure("LSCV_CUDA_TOTAL",totalEnd-totalStart);
	
	return hmin;
}
