#ifndef MATRICES_H
#define MATRICES_H

#include "utils.h"
#include <memory.h>
#include <math.h>
#include <stdio.h>
#include "alglib/ap.h"
#include "alglib/linalg.h"


float* transposeMatrix(float* m, unsigned int rows, unsigned int cols);

/**
 * \brief Creates a covariance matrix based on the given matrix.
 * 
 * Creates a covariance matrix based on the given matrix. The resulting covariance matrix is stored in a new array 
 * allocated by this function. Remember to free them later!
 * \param m a one dimensional array containing row-major aligned input matrix
 * \param rows number of input matrix rows
 * \param cols number of input matrix columns
 * \return a pointer to the newly allocated array containing output (row-major) matrix
 */
float* covarianceMatrix(float* m, uint rows, uint cols);

/**
 * \brief Performs a LUP matrix decomposition of a square matrix.
 * 
 * Performs a LUP matrix decomposition of the input square matrix (partial pivoting). The resulting matrices are stored in new arrays
 * allocated by this function. Remember to free them later!
 * 
 * \param m a one dimensional array containing row-major aligned input matrix
 * \param n the number of rows/columns in the input matrix
 * \param o a parameter returning the pointer to the array storing L and U matrices (L and U are lower and upper diagonal, and fit in one square array).
 * \param p a parameter returning the pointer to the array storing permutation matrix P
 * \param c a parameter returning the number of column switches (has influence on the determinant sign)
 */
void lupMatrixDecomposition(float *m, uint n, float **_o, int **_p, int* c);

/**
 * \brief An auxiliary function inverting a lower trangular matrix.
 * 
 * Inverts a lower triangular matrix. Auxiliary function used by the generic matrix inversion procedure.
 * 
 * \param n the number of rows/columns in the input matrix
 * \param m row major aligned input matrix
 * \param o row major aligned pointer to the output matrix (the output matrix must be preallocated)
 */
void invertLowerTriangularMatrixWithoutAlloc(uint n, float *m, float *o);

/**
 * \brief An auxiliary function inverting an upper trangular matrix.
 * 
 * Inverts an upper triangular matrix. Auxiliary function used by the generic matrix inversion procedure.
 * 
 * \param n the number of rows/columns in the input matrix
 * \param m row major aligned input matrix
 * \param o row major aligned pointer to the output matrix (the output matrix must be preallocated)
 */
void invertUpperTriangularMatrixWithoutAlloc(uint n, float *m, float *o);


/**
 * \brief An auxiliary function which, based on the result of the lupDecomposition finds the matrix inverse and determinant.
 * 
 * An auxiliary function which, based on the result of the lupDecomposition finds the matrix inverse and determinant.
 * The array storing the resulting matrix is allocated by this function.
 *
 * \param n the number of rows/columns in the input matrices
 * \param lu the input matrix storing L and U matrices.
 * \param p input matrix storing P permutation matrix
 * \param inv output array storing the inverted matrix
 * \param det output parameter which returns the matrix determinant
 * \param c number of column switches during LUP decomposition (has influence on the determinant sign)
 */
void invertLUPResultAndFindDeterminant(uint n, float *lu, int *p, float** inv, float *det, int c);

/**
 * \brief Inverts the input matrix and finds its determinant.
 * 
 * Main function used for inverting matrices. As a side effect it also calculates the matrix determinant.
 * 
 * \param n the number of rows/columns in the input matrix
 * \param m row major aligned input matrix
 * \param _o output array storing the inverted matrix (array is allocated by this function)
 * \param _det output parameter returning determinant of the matrix
 */
void invertMatrixAndFindDeterminant(uint n, float *m, float** _o, float *_det);

/**
 * \brief Calculates the square root of a matrix.
 * 
 * Calculates the square root of a matrix using matrix diagonalization method.
 * 
 * \param m input symmetric square matrix
 * \param n number of rows/cols of a matrix
 * \returns square root of a matrix
 */

float* sqrtMatrix(float *m, uint n);


/**
 * \brief Multiplies matrix by scalar.
 * 
 * \param m matrix
 * \param cols the number of columns in the matrix
 * \param rows the number of rows in the matrix
 * \param coeff the scalar multiplier
 */
void multMatrix(float *m,int cols, int rows,float coeff);

void multMatrix(float *out,float *m1,float *m2, int n);


#endif
