#include "plugincudakernels.h"
#include "cudamath.h"
#include <cutil_inline.h>
#include "cudautils.h"

__global__ void generateArrays(float *in, float *out1, float *out2, unsigned int n);


inline __device__ float k6d(float x) {
	//return 1.0f;
	float x2=x*x;
	return k_coeff*exp(-0.5f*x2)*(x2*(x2*(x2-15.0f)+45.0f)-15.0f);
}

inline __device__ float k4d(float x) {
	//return 1.0f;
	float x2=x*x;
	return k_coeff*exp(-0.5f*x2)*(x2*(x2-6.0f)+3.0f);
}


template <unsigned int blockSize>
__global__ void reduceK6Kernel(float *g_idata, float *g_odata, unsigned int deltaY, float g1, unsigned int n) {
    extern __shared__ float sdata[];

    unsigned int tid = threadIdx.x;
    unsigned int i = blockIdx.x*(blockSize*2) + tid;
    unsigned int gridSize = blockSize*2*gridDim.x;
		float xj=g_idata[blockIdx.y+deltaY];
    sdata[tid] = 0.0f; 
		
	  while (i+blockSize < n) { sdata[tid] += k6d((g_idata[i]-xj)/g1) + k6d((g_idata[i+blockSize]-xj)/g1);  i += gridSize;  } 
	  if (i < n ) { sdata[tid] += k6d((g_idata[i]-xj)/g1); } 
    __syncthreads();

    if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
    if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
    if (blockSize >= 128) { if (tid < 64)  { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }
    
    if (tid < 32) {
        if (blockSize >=  64) sdata[tid] += sdata[tid + 32];
        if (blockSize >=  32) sdata[tid] += sdata[tid + 16];
        if (blockSize >=  16) sdata[tid] += sdata[tid +  8];
        if (blockSize >=    8) sdata[tid] += sdata[tid +  4];
        if (blockSize >=    4) sdata[tid] += sdata[tid +  2];
        if (blockSize >=    2) sdata[tid] += sdata[tid +  1];
    }

    if (tid == 0) g_odata[blockIdx.x+gridDim.x*(blockIdx.y+deltaY)] = sdata[0];
}

template <unsigned int blockSize> __host__ float reduceK6(float *d_idata, float g1, unsigned int n) {
	unsigned int kernelRunTimes=intDivCeil(n,65535);
	unsigned int gridWidth=min(65535,intDivCeil(n,blockSize*2));
	unsigned int tempBufferSize=gridWidth*n;
	
	float *d_tempBuffer;
	
	cutilSafeCall(cudaMalloc((void**)&d_tempBuffer,sizeof(float)*tempBufferSize));
	
	for (int i=0;i<kernelRunTimes;i++) {
		dim3 grid(gridWidth,min(65535,n-i*65535));
		//printf("blocks started %d \n",grid.x*grid.y);
		reduceK6Kernel<blockSize><<<grid,blockSize,blockSize*sizeof(float)>>>(d_idata, d_tempBuffer, i*65535, g1, n);
	}
	
	//cutilSafeCall(cudaThreadSynchronize());
	
	float result=lossyReduce<blockSize>(d_tempBuffer,tempBufferSize);
	
	cutilSafeCall(cudaFree(d_tempBuffer));
	
	return result;
}

template <unsigned int blockSize>
__global__ void reduceK4Kernel(float *g_idata, float *g_odata, unsigned int deltaY, float g2, unsigned int n) {
    extern __shared__ float sdata[];

    unsigned int tid = threadIdx.x;
    unsigned int i = blockIdx.x*(blockSize*2) + tid;
    unsigned int gridSize = blockSize*2*gridDim.x;
		float xj=g_idata[blockIdx.y+deltaY];
    sdata[tid] = 0.0f; 
		
	  while (i+blockSize < n) { sdata[tid] += k4d((g_idata[i]-xj)/g2) + k4d((g_idata[i+blockSize]-xj)/g2);  i += gridSize;  } 
	  if (i < n ) { sdata[tid] += k4d((g_idata[i]-xj)/g2); } 
    __syncthreads();

    if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
    if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
    if (blockSize >= 128) { if (tid < 64)  { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }
    
    if (tid < 32) {
        if (blockSize >=  64) sdata[tid] += sdata[tid + 32];
        if (blockSize >=  32) sdata[tid] += sdata[tid + 16];
        if (blockSize >=  16) sdata[tid] += sdata[tid +  8];
        if (blockSize >=    8) sdata[tid] += sdata[tid +  4];
        if (blockSize >=    4) sdata[tid] += sdata[tid +  2];
        if (blockSize >=    2) sdata[tid] += sdata[tid +  1];
    }

    if (tid == 0) g_odata[blockIdx.x+gridDim.x*(blockIdx.y+deltaY)] = sdata[0];
}

template <unsigned int blockSize> __host__ float reduceK4(float *d_idata, float g2, unsigned int n) {
	unsigned int kernelRunTimes=intDivCeil(n,65535);
	unsigned int gridWidth=min(65535,intDivCeil(n,blockSize*2));
	unsigned int tempBufferSize=gridWidth*n;
	
	float *d_tempBuffer;
	
	cutilSafeCall(cudaMalloc((void**)&d_tempBuffer,sizeof(float)*tempBufferSize));
	
	for (int i=0;i<kernelRunTimes;i++) {
		dim3 grid(gridWidth,min(65535,n-i*65535));
		reduceK4Kernel<blockSize><<<grid,blockSize,blockSize*sizeof(float)>>>(d_idata, d_tempBuffer, i*65535, g2, n);
	}
	
	//cutilSafeCall(cudaThreadSynchronize());
	
	float result=lossyReduce<blockSize>(d_tempBuffer,tempBufferSize);
	
	cutilSafeCall(cudaFree(d_tempBuffer));
	
	return result;
}


__global__ void generateArrays(float *in, float *out1, float *out2, uint n) {
	uint i=threadIdx.x+blockIdx.x*blockDim.x;
	
	if (i<n) {
		float x=in[i];
		out1[i]=x;
		out2[i]=x*x;
	}
}

__device__ inline float ceilSquareRoot(unsigned int x) {
	float a=sqrtf(x);
	//return round(a)*round(a)==x?round(a):ceil(a);
	return round(a)*round(a)==x?round(a):a;
}


//TODO jeden wątek 4 elementy tablicy
/*template <int blockSize,int side> __global__ void aggK6values(float *data, float* out, int n,float g1) {
	extern __shared__ float base[];

	float* Mi=base;
	float* Mj=base+side;
	float* sdata=base+2*side;

	int bx=blockIdx.x+blockIdx.y*gridDim.x;

	int i=ceil((ceilSquareRoot((bx<<3)+9)-3.0f)/2.0f);
	int j=bx-((i*(i+1))>>1);

	int tx=threadIdx.x%side;
	int ty=threadIdx.x/side;

	int iBegin=i*side;
	int jBegin=j*side;

	if (ty==0) {
		Mj[tx]=data[jBegin+tx]; 
		Mi[tx]=data[iBegin+tx]; 
	}
	
	sdata[threadIdx.x]=0.0f;

	__syncthreads();

	if ((jBegin+tx)<(iBegin+ty)) {
		if (((jBegin+tx)<n) && ((iBegin+ty)<n) ) {
			sdata[ty*side+tx]=k6d((Mi[ty]-Mj[tx])/g1);
		} 
	}
	
	__syncthreads();
	
	if (blockSize >= 512) { if (threadIdx.x < 256) { sdata[threadIdx.x] += sdata[threadIdx.x + 256]; } __syncthreads(); }
	if (blockSize >= 256) { if (threadIdx.x < 128) { sdata[threadIdx.x] += sdata[threadIdx.x + 128]; } __syncthreads(); }
	if (blockSize >= 128) { if (threadIdx.x < 64)  { sdata[threadIdx.x] += sdata[threadIdx.x +  64]; } __syncthreads(); }
    
	if (threadIdx.x < 32) {
		volatile float* smem = sdata;
		if (blockSize >=  64) smem[threadIdx.x] += smem[threadIdx.x + 32];
		if (blockSize >=  32) smem[threadIdx.x] += smem[threadIdx.x + 16];
		if (blockSize >=  16) smem[threadIdx.x] += smem[threadIdx.x +  8];
		if (blockSize >=   8) smem[threadIdx.x] += smem[threadIdx.x +  4];
		if (blockSize >=   4) smem[threadIdx.x] += smem[threadIdx.x +  2];
		if (blockSize >=   2) smem[threadIdx.x] += smem[threadIdx.x +  1];
	}
	
	
	//out[threadIdx.x]=sdata[threadIdx.x];
	
	__syncthreads();

	if (threadIdx.x==0) {
		if ((jBegin+tx)<n && (iBegin+ty)<n ) {	
			volatile float* smem = sdata;
			out[bx]=smem[0];
		}
	}
}




//WARNING blockSize w zasadzie poprawny tylko dla 256 i side =16 side*side musi byc rowne blocksize
template<int blockSize,int side> float calcSumK6(float *d_x, float g1, int  n) {
	int blocks=(n/16)*((n/16)+1)/2;
	int blocksY=intDivCeil(blocks,65535);
	int blocksX=intDivCeil(blocks,blocksY);
	
	float *d_tempx;
	cutilSafeCall(cudaMalloc((void**)&d_tempx,sizeof(float)*blocksX*blocksY));
	
	dim3 grid(blocksX,blocksY);
	//printf("grid %d %d \n",grid.x,grid.y);
	//printf("blocks started %d \n",blocks);
	
	aggK6values<blockSize,side><<<grid,blockSize,(blockSize+2*side)*sizeof(float)>>>(d_x, d_tempx, n,g1);
	
	
	
	float result=lossyReduce<512>(d_tempx, blocks);
	
	cutilSafeCall(cudaFree(d_tempx));
	
	return result;
}*/

template <int blockSize,int side> __global__ void aggK6values(float *data, float* out, int n,float g1) {
	extern __shared__ float base[];

	float* Mi=base;
	float* Mj=base+side*2;
	float* sdata=base+4*side;

	int bx=blockIdx.x+blockIdx.y*gridDim.x;

	int i=ceil((ceilSquareRoot((bx<<3)+9)-3.0f)/2.0f);
	int j=bx-((i*(i+1))>>1);

	int tx=threadIdx.x%side;
	int ty=threadIdx.x/side;

	int iBegin=i*side*2;
	int jBegin=j*side*2;

	//TODO wydajniejsze przepisanie danych
	
	if (ty<=1) {
		Mj[tx+ty*side]=data[jBegin+tx+ty*side]; 
		Mi[tx+ty*side]=data[iBegin+tx+ty*side]; 
	}
	
	/*if (ty==0) {
		Mj[tx]=data[jBegin+tx]; 
		Mi[tx]=data[iBegin+tx]; 
	} if (ty==1) {
		Mj[tx+side/2]=data[jBegin+tx+side/2];
		Mi[tx+side/2]=data[iBegin+tx+side/2];
	}*/

	__syncthreads();
	
	float res;
	res=k6d((Mi[ty]-Mj[tx])/g1)*((jBegin+tx)<(iBegin+ty));
	res+=k6d((Mi[ty]-Mj[tx+side])/g1)*((jBegin+tx+side)<(iBegin+ty));
	res+=k6d((Mi[ty+side]-Mj[tx])/g1)*((jBegin+tx)<(iBegin+ty+side));
	res+=k6d((Mi[ty+side]-Mj[tx+side])/g1)*((jBegin+tx+side)<(iBegin+ty+side));
	
	/*
	res=((jBegin+tx)<(iBegin+ty));
	res+=((jBegin+tx+side/2)<(iBegin+ty));
	res+=((jBegin+tx)<(iBegin+ty+side/2));
	res+=((jBegin+tx+side/2)<(iBegin+ty+side/2));*/
	
	sdata[ty*side+tx]=res;
	
	__syncthreads();
	
	if (blockSize >= 512) { if (threadIdx.x < 256) { sdata[threadIdx.x] += sdata[threadIdx.x + 256]; } __syncthreads(); }
	if (blockSize >= 256) { if (threadIdx.x < 128) { sdata[threadIdx.x] += sdata[threadIdx.x + 128]; } __syncthreads(); }
	if (blockSize >= 128) { if (threadIdx.x < 64)  { sdata[threadIdx.x] += sdata[threadIdx.x +  64]; } __syncthreads(); }
    
	if (threadIdx.x < 32) {
		volatile float* smem = sdata;
		if (blockSize >=  64) smem[threadIdx.x] += smem[threadIdx.x + 32];
		if (blockSize >=  32) smem[threadIdx.x] += smem[threadIdx.x + 16];
		if (blockSize >=  16) smem[threadIdx.x] += smem[threadIdx.x +  8];
		if (blockSize >=   8) smem[threadIdx.x] += smem[threadIdx.x +  4];
		if (blockSize >=   4) smem[threadIdx.x] += smem[threadIdx.x +  2];
		if (blockSize >=   2) smem[threadIdx.x] += smem[threadIdx.x +  1];
	}
	
	
	__syncthreads();

	
	if (threadIdx.x==0) {
		//if ((jBegin+tx)<n && (iBegin+ty)<n ) {	
			volatile float* smem = sdata;
			out[bx]=smem[0];
		//}
	}
	
}

template<int blockSize,int side> float calcSumK6(float *d_x, float g1, int  n) {
	int blocks=(n/side/2)*((n/side/2)+1)/2;
	int blocksY=intDivCeil(blocks,65535);
	int blocksX=intDivCeil(blocks,blocksY);
	
	float *d_tempx;
	cutilSafeCall(cudaMalloc((void**)&d_tempx,sizeof(float)*blocksX*blocksY));
	
	dim3 grid(blocksX,blocksY);
	
	aggK6values<blockSize,side><<<grid,blockSize,(blockSize+4*side)*sizeof(float)>>>(d_x, d_tempx, n,g1);
	
	/*float *tempx=new float[sizeof(float)*blocksX*blocksY];
	cutilSafeCall(cudaMemcpy(tempx,d_tempx,sizeof(float)*blocksX*blocksY,cudaMemcpyDeviceToHost));
	
	float s=0;
	for (int i=0;i<blocks;i++) {
		printf("%.0f ",tempx[i]);
		s+=tempx[i];
	}
	printf("total %f \n",s);*/
	
	float result=lossyReduce<512>(d_tempx, blocks);
	
	cutilSafeCall(cudaFree(d_tempx));
	
	return result;
}

/*template <int blockSize,int side> __global__ void aggK4values(float *data, float* out, int n,float g2) {
	extern __shared__ float base[];

	float* Mi=base;
	float* Mj=base+side;
	float* sdata=base+2*side;

	int bx=blockIdx.x+blockIdx.y*gridDim.x;

	int i=ceil((ceilSquareRoot((bx<<3)+9)-3.0f)/2.0f);
	int j=bx-((i*(i+1))>>1);

	int tx=threadIdx.x%side;
	int ty=threadIdx.x/side;

	int iBegin=i*side;
	int jBegin=j*side;

	if (ty==0) {
		Mj[tx]=data[jBegin+tx]; 
		Mi[tx]=data[iBegin+tx]; 
	}
	
	sdata[threadIdx.x]=0.0f;

	__syncthreads();

	if ((jBegin+tx)<(iBegin+ty)) {
		if (((jBegin+tx)<n) && ((iBegin+ty)<n) ) {
			sdata[ty*side+tx]=k4d((Mi[ty]-Mj[tx])/g2);
		} 
	}
	
	__syncthreads();
	
	if (blockSize >= 512) { if (threadIdx.x < 256) { sdata[threadIdx.x] += sdata[threadIdx.x + 256]; } __syncthreads(); }
	if (blockSize >= 256) { if (threadIdx.x < 128) { sdata[threadIdx.x] += sdata[threadIdx.x + 128]; } __syncthreads(); }
	if (blockSize >= 128) { if (threadIdx.x < 64)  { sdata[threadIdx.x] += sdata[threadIdx.x +  64]; } __syncthreads(); }
    
	if (threadIdx.x < 32) {
		volatile float* smem = sdata;
		if (blockSize >=  64) smem[threadIdx.x] += smem[threadIdx.x + 32];
		if (blockSize >=  32) smem[threadIdx.x] += smem[threadIdx.x + 16];
		if (blockSize >=  16) smem[threadIdx.x] += smem[threadIdx.x +  8];
		if (blockSize >=   8) smem[threadIdx.x] += smem[threadIdx.x +  4];
		if (blockSize >=   4) smem[threadIdx.x] += smem[threadIdx.x +  2];
		if (blockSize >=   2) smem[threadIdx.x] += smem[threadIdx.x +  1];
	}
	
	//out[threadIdx.x]=sdata[threadIdx.x];
	
	__syncthreads();

	if ((jBegin+tx)<n && (iBegin+ty)<n ) {	
		if (threadIdx.x==0) {
			volatile float* smem = sdata;
			out[bx]=smem[0];
		}
	}
}


//WARNING blockSize w zasadzie poprawny tylko dla 256 i side =16 side*side musi byc rowne blocksize
template<int blockSize,int side> float calcSumK4(float *d_x, float g2, int  n) {
	int blocks=(n/16)*((n/16)+1)/2;
	int blocksY=intDivCeil(blocks,65535);
	int blocksX=intDivCeil(blocks,blocksY);
	
	float *d_tempx;
	cutilSafeCall(cudaMalloc((void**)&d_tempx,sizeof(float)*blocksX*blocksY));
	
	dim3 grid(blocksX,blocksY);
	//printf("grid %d %d \n",grid.x,grid.y);
	
	aggK4values<blockSize,side><<<grid,blockSize,(blockSize+2*side)*sizeof(float)>>>(d_x, d_tempx, n,g2);
	
	
	float result=lossyReduce<512>(d_tempx, blocks);
	
	cutilSafeCall(cudaFree(d_tempx));
	
	return result;
}*/

template <int blockSize,int side> __global__ void aggK4values(float *data, float* out, int n,float g2) {
		extern __shared__ float base[];

	float* Mi=base;
	float* Mj=base+side*2;
	float* sdata=base+4*side;

	int bx=blockIdx.x+blockIdx.y*gridDim.x;

	int i=ceil((ceilSquareRoot((bx<<3)+9)-3.0f)/2.0f);
	int j=bx-((i*(i+1))>>1);

	int tx=threadIdx.x%side;
	int ty=threadIdx.x/side;

	int iBegin=i*side*2;
	int jBegin=j*side*2;

	//TODO wydajniejsze przepisanie danych
	
	if (ty<=1) {
		Mj[tx+ty*side]=data[jBegin+tx+ty*side]; 
		Mi[tx+ty*side]=data[iBegin+tx+ty*side]; 
	}
	
	/*if (ty==0) {
		Mj[tx]=data[jBegin+tx]; 
		Mi[tx]=data[iBegin+tx]; 
	} if (ty==1) {
		Mj[tx+side/2]=data[jBegin+tx+side/2];
		Mi[tx+side/2]=data[iBegin+tx+side/2];
	}*/

	__syncthreads();
	
	float res;
	res=k4d((Mi[ty]-Mj[tx])/g2)*((jBegin+tx)<(iBegin+ty));
	res+=k4d((Mi[ty]-Mj[tx+side])/g2)*((jBegin+tx+side)<(iBegin+ty));
	res+=k4d((Mi[ty+side]-Mj[tx])/g2)*((jBegin+tx)<(iBegin+ty+side));
	res+=k4d((Mi[ty+side]-Mj[tx+side])/g2)*((jBegin+tx+side)<(iBegin+ty+side));
	
	/*
	res=((jBegin+tx)<(iBegin+ty));
	res+=((jBegin+tx+side/2)<(iBegin+ty));
	res+=((jBegin+tx)<(iBegin+ty+side/2));
	res+=((jBegin+tx+side/2)<(iBegin+ty+side/2));*/
	
	sdata[ty*side+tx]=res;
	
	__syncthreads();
	
	if (blockSize >= 512) { if (threadIdx.x < 256) { sdata[threadIdx.x] += sdata[threadIdx.x + 256]; } __syncthreads(); }
	if (blockSize >= 256) { if (threadIdx.x < 128) { sdata[threadIdx.x] += sdata[threadIdx.x + 128]; } __syncthreads(); }
	if (blockSize >= 128) { if (threadIdx.x < 64)  { sdata[threadIdx.x] += sdata[threadIdx.x +  64]; } __syncthreads(); }
    
	if (threadIdx.x < 32) {
		volatile float* smem = sdata;
		if (blockSize >=  64) smem[threadIdx.x] += smem[threadIdx.x + 32];
		if (blockSize >=  32) smem[threadIdx.x] += smem[threadIdx.x + 16];
		if (blockSize >=  16) smem[threadIdx.x] += smem[threadIdx.x +  8];
		if (blockSize >=   8) smem[threadIdx.x] += smem[threadIdx.x +  4];
		if (blockSize >=   4) smem[threadIdx.x] += smem[threadIdx.x +  2];
		if (blockSize >=   2) smem[threadIdx.x] += smem[threadIdx.x +  1];
	}
	
	
	__syncthreads();

	
	if (threadIdx.x==0) {
		//if ((jBegin+tx)<n && (iBegin+ty)<n ) {	
			volatile float* smem = sdata;
			out[bx]=smem[0];
		//}
	}
}


//WARNING blockSize w zasadzie poprawny tylko dla 256 i side =16 side*side musi byc rowne blocksize
template<int blockSize,int side> float calcSumK4(float *d_x, float g2, int  n) {
	int blocks=(n/side/2)*((n/side/2)+1)/2;
	int blocksY=intDivCeil(blocks,65535);
	int blocksX=intDivCeil(blocks,blocksY);
	
	float *d_tempx;
	cutilSafeCall(cudaMalloc((void**)&d_tempx,sizeof(float)*blocksX*blocksY));
	
	dim3 grid(blocksX,blocksY);
	//printf("grid %d %d \n",grid.x,grid.y);
	
	aggK4values<blockSize,side><<<grid,blockSize,(blockSize+4*side)*sizeof(float)>>>(d_x, d_tempx, n,g2);
	
	/*float *tempx=new float[sizeof(float)*blocksX*blocksY];
	cutilSafeCall(cudaMemcpy(tempx,d_tempx,sizeof(float)*blocksX*blocksY,cudaMemcpyDeviceToHost));
	
	float s=0;
	for (int i=0;i<blocks;i++) {
		printf("%.0f ",tempx[i]);
		s+=tempx[i];
	}
	printf("total %f \n",s);*/
	
	float result=lossyReduce<512>(d_tempx, blocks);
	
	cutilSafeCall(cudaFree(d_tempx));
	
	return result;
}



float calcH(HAlgorithm<float>* a,float *x, int n) {
	float *d_x;
	float *d_tempx;
	float *d_x2;
	//int n=getN();
	//float *x=getData();
	
	TIMESTAMP start,end,total, totalParallelizable,tempStart;
	
	totalParallelizable=0;
	
	start=getTimer();
	cutilSafeCall(cudaMalloc((void**)&d_x,sizeof(float)*n));
	cutilSafeCall(cudaMalloc((void**)&d_tempx,sizeof(float)*n));
	cutilSafeCall(cudaMalloc((void**)&d_x2,sizeof(float)*n));
	cutilSafeCall(cudaMemcpy(d_x,x,sizeof(float)*n,cudaMemcpyHostToDevice));
	end=getTimer();
	total=end-start;
	a->addMeasure("PLUGIN_CUDA_TRANSFER",end-start);
	
	
	start=getTimer();
	tempStart=start;
	generateArrays<<<intDivCeil(n,512),512>>>(d_x,d_tempx,d_x2,n);
	
	float sumx=lossyReduce<512>(d_tempx,n);
	float sumx2=lossyReduce<512>(d_x2,n);
	
	totalParallelizable+=getTimer()-tempStart;
	
	cutilSafeCall(cudaFree(d_x2));
	cutilSafeCall(cudaFree(d_tempx));
	
	float var=estVar(n,sumx,sumx2);
	float stdDev=estStdDev(var);
	float _psi8=psi8(stdDev);
	float g1=f4(_psi8,n);
	
	tempStart=getTimer();
	float sumk6=calcSumK6<256,16>(d_x, g1, n);
	totalParallelizable+=getTimer()-tempStart;
	

	float _psi6=psi6(g1,n,sumk6);
	float g2=f2(_psi6,n);
	
	tempStart=getTimer();
	float sumk4=calcSumK4<256,16>(d_x, g2, n);
	totalParallelizable+=getTimer()-tempStart;
	
	cutilSafeCall(cudaFree(d_x));
	
	
	float _psi4=psi4(g2,n,sumk4);
	float h=hcoeff(_psi4,n);
	end=getTimer();
	total+=end-start;
	a->addMeasure("PLUGIN_CUDA_CALC",end-start);
	a->addMeasure("PLUGIN_CUDA_TOTAL",total);
	a->addMeasure("PLUGIN_CUDA_PARALLELIZABLE",totalParallelizable);
	
	return h;
}